from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, List, Optional, Set, Tuple

from netqasm.lang.instr import core
from netqasm.lang.operand import Template

from qoala.lang.ehi import EhiNetworkInfo, EhiNodeInfo
from qoala.lang.hostlang import (
    BasicBlock,
    BasicBlockType,
    ReceiveCMsgOp,
    RunRequestOp,
    RunSubroutineOp,
)
from qoala.lang.program import QoalaProgram
from qoala.lang.request import CallbackType
from qoala.lang.routine import LocalRoutine
from qoala.runtime.program import ProgramInstance


class ProcessorType(Enum):
    CPU = 0
    QPU = auto()


class QoalaTask:
    def __init__(
        self,
        task_id: int,
        processor_type: ProcessorType,
        pid: int,
        duration: Optional[float] = None,
    ) -> None:
        self._task_id = task_id
        self._processor_type = processor_type
        self._pid = pid
        self._duration = duration

    def __str__(self) -> str:
        s = f"{self.__class__.__name__}(pid={self.pid}, tid={self.task_id})"
        if not self.is_epr_task() and hasattr(self, "block_name"):
            s += f"block={self.block_name}"  # type: ignore
        return s

    @property
    def task_id(self) -> int:
        return self._task_id

    @property
    def processor_type(self) -> ProcessorType:
        return self._processor_type

    @property
    def pid(self) -> int:
        return self._pid

    @property
    def duration(self) -> Optional[float]:
        return self._duration

    def is_epr_task(self) -> bool:
        return isinstance(self, SinglePairTask) or isinstance(self, MultiPairTask)

    def is_event_task(self) -> bool:
        return isinstance(self, HostEventTask)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QoalaTask):
            return NotImplemented
        return (
            self.task_id == other.task_id
            and self.processor_type == other.processor_type
            and self.pid == other.pid
            and self.duration == other.duration
        )


class HostLocalTask(QoalaTask):
    def __init__(
        self,
        task_id: int,
        pid: int,
        block_name: str,
        duration: Optional[float] = None,
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.CPU,
            pid=pid,
            duration=duration,
        )
        self._block_name = block_name

    @property
    def block_name(self) -> str:
        return self._block_name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HostLocalTask):
            return NotImplemented
        return super().__eq__(other) and self.block_name == other.block_name


class HostEventTask(QoalaTask):
    def __init__(
        self, task_id: int, pid: int, block_name: str, duration: Optional[float] = None
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.CPU,
            pid=pid,
            duration=duration,
        )
        self._block_name = block_name

    @property
    def block_name(self) -> str:
        return self._block_name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HostEventTask):
            return NotImplemented
        return super().__eq__(other) and self.block_name == other.block_name


class LocalRoutineTask(QoalaTask):
    def __init__(
        self,
        task_id: int,
        pid: int,
        block_name: str,
        shared_ptr: int,
        duration: Optional[float] = None,
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.QPU,
            pid=pid,
            duration=duration,
        )
        self._block_name = block_name
        self._shared_ptr = shared_ptr

    @property
    def block_name(self) -> str:
        return self._block_name

    @property
    def shared_ptr(self) -> int:
        return self._shared_ptr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LocalRoutineTask):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.block_name == other.block_name
            and self.shared_ptr == self.shared_ptr
        )


class PreCallTask(QoalaTask):
    def __init__(
        self,
        task_id: int,
        pid: int,
        block_name: str,
        shared_ptr: int,  # used to identify shared (with other tasks) lrcall/rrcall objects
        duration: Optional[float] = None,
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.CPU,
            pid=pid,
            duration=duration,
        )
        self._block_name = block_name
        self._shared_ptr = shared_ptr

    @property
    def block_name(self) -> str:
        return self._block_name

    @property
    def shared_ptr(self) -> int:
        return self._shared_ptr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PreCallTask):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.block_name == other.block_name
            and self.shared_ptr == self.shared_ptr
        )


class PostCallTask(QoalaTask):
    def __init__(
        self,
        task_id: int,
        pid: int,
        block_name: str,
        shared_ptr: int,  # used to identify shared (with other tasks) lrcall/rrcall objects
        duration: Optional[float] = None,
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.CPU,
            pid=pid,
            duration=duration,
        )
        self._block_name = block_name
        self._shared_ptr = shared_ptr

    @property
    def block_name(self) -> str:
        return self._block_name

    @property
    def shared_ptr(self) -> int:
        return self._shared_ptr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PostCallTask):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.block_name == other.block_name
            and self.shared_ptr == other.shared_ptr
        )


class SinglePairTask(QoalaTask):
    def __init__(
        self,
        task_id: int,
        pid: int,
        pair_index: int,
        shared_ptr: int,  # used to identify shared (with other tasks) lrcall/rrcall objects
        duration: Optional[float] = None,
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.QPU,
            pid=pid,
            duration=duration,
        )
        self._pair_index = pair_index
        self._shared_ptr = shared_ptr

    @property
    def pair_index(self) -> int:
        return self._pair_index

    @property
    def shared_ptr(self) -> int:
        return self._shared_ptr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SinglePairTask):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.pair_index == other.pair_index
            and self.shared_ptr == other.shared_ptr
        )


class MultiPairTask(QoalaTask):
    def __init__(
        self,
        task_id: int,
        pid: int,
        shared_ptr: int,  # used to identify shared (with other tasks) lrcall/rrcall objects
        duration: Optional[float] = None,
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.QPU,
            pid=pid,
            duration=duration,
        )
        self._shared_ptr = shared_ptr

    @property
    def shared_ptr(self) -> int:
        return self._shared_ptr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MultiPairTask):
            return NotImplemented
        return super().__eq__(other) and self.shared_ptr == other.shared_ptr


class SinglePairCallbackTask(QoalaTask):
    def __init__(
        self,
        task_id: int,
        pid: int,
        callback_name: str,
        pair_index: int,
        shared_ptr: int,  # used to identify shared (with other tasks) lrcall/rrcall objects
        duration: Optional[float] = None,
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.QPU,
            pid=pid,
            duration=duration,
        )
        self._callback_name = callback_name
        self._pair_index = pair_index
        self._shared_ptr = shared_ptr

    @property
    def callback_name(self) -> str:
        return self._callback_name

    @property
    def pair_index(self) -> int:
        return self._pair_index

    @property
    def shared_ptr(self) -> int:
        return self._shared_ptr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SinglePairCallbackTask):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.callback_name == other.callback_name
            and self.pair_index == other.pair_index
            and self.shared_ptr == other.shared_ptr
        )


class MultiPairCallbackTask(QoalaTask):
    def __init__(
        self,
        task_id: int,
        pid: int,
        callback_name: str,
        shared_ptr: int,  # used to identify shared (with other tasks) lrcall/rrcall objects
        duration: Optional[float] = None,
    ) -> None:
        super().__init__(
            task_id=task_id,
            processor_type=ProcessorType.QPU,
            pid=pid,
            duration=duration,
        )
        self._callback_name = callback_name
        self._shared_ptr = shared_ptr

    @property
    def callback_name(self) -> str:
        return self._callback_name

    @property
    def shared_ptr(self) -> int:
        return self._shared_ptr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MultiPairCallbackTask):
            return NotImplemented
        return (
            super().__eq__(other)
            and self.callback_name == other.callback_name
            and self.shared_ptr == other.shared_ptr
        )


@dataclass
class TaskInfo:
    task: QoalaTask
    predecessors: Set[int]
    ext_predecessors: Set[int]
    successors: Set[int]
    deadline: Optional[int]
    rel_deadlines: Dict[int, int]
    ext_rel_deadlines: Dict[int, int]
    start_time: Optional[float]
    deadline_set: bool = False

    @classmethod
    def only_task(cls, task: QoalaTask) -> TaskInfo:
        return TaskInfo(task, set(), set(), set(), None, {}, {}, None)

    def is_cpu_task(self) -> bool:
        return self.task.processor_type == ProcessorType.CPU

    def is_qpu_task(self) -> bool:
        return self.task.processor_type == ProcessorType.QPU


@dataclass
class TaskGraph:
    """DAG of Tasks."""

    def __init__(self, tasks: Optional[Dict[int, TaskInfo]] = None) -> None:
        if tasks is None:
            self._tasks: Dict[int, TaskInfo] = {}
        else:
            self._tasks = tasks

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TaskGraph):
            raise NotImplementedError
        return self._tasks == other._tasks

    def __str__(self) -> str:
        return "\n".join(f"{i}: {t}" for i, t in self._tasks.items())

    def add_tasks(self, tasks: List[QoalaTask]) -> None:
        for task in tasks:
            self._tasks[task.task_id] = TaskInfo.only_task(task)

    def add_precedences(self, precedences: List[Tuple[int, int]]) -> None:
        # an entry (x, y) means that x precedes y (y should execute after x)
        for (x, y) in precedences:
            assert x in self._tasks and y in self._tasks
            self._tasks[y].predecessors.add(x)
            self._tasks[x].successors.add(y)

    def update_successors(self) -> None:
        # Make sure all `successors` of all tinfos match all predecessors
        for tid, tinfo in self.get_tasks().items():
            for pred in tinfo.predecessors:
                pred_tinfo = self.get_tinfo(pred)
                if tid not in pred_tinfo.successors:
                    pred_tinfo.successors.add(tid)

    def add_ext_precedences(self, precedences: List[Tuple[int, int]]) -> None:
        # an entry (x, y) means that x (which is not in this graph) precedes y
        # (which is in this graph)
        for (x, y) in precedences:
            assert x not in self._tasks and y in self._tasks
            self._tasks[y].ext_predecessors.add(x)

    def add_deadlines(self, deadlines: List[Tuple[int, int]]) -> None:
        for (x, d) in deadlines:
            assert x in self._tasks
            self._tasks[x].deadline = d

    def add_rel_deadlines(self, deadlines: List[Tuple[Tuple[int, int], int]]) -> None:
        # entry ((x, y), d) means
        # task y must start at most time d time units after task x has finished
        for ((x, y), d) in deadlines:
            assert x in self._tasks and y in self._tasks
            self._tasks[y].rel_deadlines[x] = d

    def add_ext_rel_deadlines(
        self, deadlines: List[Tuple[Tuple[int, int], int]]
    ) -> None:
        # entry ((x, y), d) means
        # task y must start at most time d time units after task x has finished
        for ((x, y), d) in deadlines:
            assert x not in self._tasks and y in self._tasks  # x is external
            self._tasks[y].ext_rel_deadlines[x] = d

    def get_tasks(self) -> Dict[int, TaskInfo]:
        return self._tasks

    def get_tinfo(self, id: int) -> TaskInfo:
        assert id in self._tasks
        return self._tasks[id]

    def task_exists_for_pid(self, pid: int) -> bool:
        for tid, tinfo in self._tasks.items():
            if tinfo.task.pid == pid:
                return True
        return False

    def get_roots(self, ignore_external: bool = False) -> List[int]:
        # Return all (IDs of) tasks that have no predecessors

        if ignore_external:
            return [
                i for i, tinfo in self._tasks.items() if len(tinfo.predecessors) == 0
            ]
        else:
            return [
                i
                for i, tinfo in self._tasks.items()
                if len(tinfo.predecessors) == 0 and len(tinfo.ext_predecessors) == 0
            ]

    def get_tasks_blocked_only_on_external(self) -> List[int]:
        return [
            i
            for i, tinfo in self._tasks.items()
            if len(tinfo.predecessors) == 0 and len(tinfo.ext_predecessors) > 0
        ]

    def get_epr_roots(self, ignore_external: bool = False) -> List[int]:
        roots = self.get_roots(ignore_external)
        return [r for r in roots if self.get_tinfo(r).task.is_epr_task()]

    def get_event_roots(self, ignore_external: bool = False) -> List[int]:
        roots = self.get_roots(ignore_external)
        return [r for r in roots if self.get_tinfo(r).task.is_event_task()]

    def linearize(self) -> List[int]:
        # Returns None if not linear
        if len(self.get_tasks()) == 0:
            return []  # empty graph is linear

        roots = self.get_roots()
        if len(roots) != 1:
            raise RuntimeError("Task Graph cannot be linearized")

        chain: List[int] = [roots[0]]
        for _ in range(len(self._tasks) - 1):
            successors = self.get_tinfo(chain[-1]).successors
            if len(successors) != 1:
                raise RuntimeError("Task Graph cannot be Linearized")
            successor = successors.pop()
            chain.append(successor)
            successors.add(successor)
        return chain

    def remove_task(self, id: int) -> None:
        assert id in self.get_roots(ignore_external=True)
        tinfo = self._tasks.pop(id)

        # Remove precedences of successor tasks
        for succ in tinfo.successors:
            succ_info = self.get_tinfo(succ)
            assert id in succ_info.predecessors
            succ_info.predecessors.remove(id)

        # Change relative deadlines to absolute ones
        for t in self._tasks.values():
            if id in t.rel_deadlines:
                t.deadline = t.rel_deadlines.pop(id)

    def decrease_deadlines(self, amount: int) -> None:
        for tinfo in self._tasks.values():
            if tinfo.deadline is not None:
                tinfo.deadline -= amount

    def get_cpu_graph(self) -> TaskGraph:
        return self.partial_graph(ProcessorType.CPU)

    def get_qpu_graph(self) -> TaskGraph:
        return self.partial_graph(ProcessorType.QPU)

    def cross_predecessors(self, task_id: int, immediate: bool = True) -> Set[int]:
        # Return all (IDs of) tasks that are predecessors that run on
        # the other processor (CPU/QPU).
        # If immediate = False, return all closest such predecessor, even if they are
        # no immediate parents.
        # If immediate = True, return only immediate parents with a different processor
        # type.
        # TODO: remove items from result set when they are ancestors of other items
        # in the set (in which case they are redundant)
        proc_type = self.get_tinfo(task_id).task.processor_type
        cross_preds = set()

        for pred in self.get_tinfo(task_id).predecessors:
            pred_type = self.get_tinfo(pred).task.processor_type
            if pred_type != proc_type:
                cross_preds.add(pred)  # immediate parent of different type
            elif not immediate:
                cross_preds = cross_preds.union(
                    self.cross_predecessors(pred, immediate)
                )
        return cross_preds

    def double_cross_predecessors(self, task_id: int) -> Set[int]:
        # Return all (IDs of) tasks that are the closest predecessors that run on
        # the same processor (CPU/QPU) but where there are tasks of the other processor
        # type inbetween (in the precedence chain).

        # For the first step: only check immediate parents that have different type.
        # Parents with same type already induce a normal precedence constraint in the
        # partial graph.
        cross_preds = self.cross_predecessors(task_id, immediate=True)
        double_cross_preds: Set[int] = set()
        for cp in cross_preds:
            # For each different-type parent, find the nearest ancestor of the original
            # type.
            double_cross_preds = double_cross_preds.union(
                self.cross_predecessors(cp, immediate=False)
            )
        return double_cross_preds

    def partial_graph(self, proc_type: ProcessorType) -> TaskGraph:
        # Filter tasks with the correct type.
        partial_tasks: Dict[int, TaskInfo] = {
            i: deepcopy(tinfo)
            for i, tinfo in self._tasks.items()
            if tinfo.task.processor_type == proc_type
        }

        # Precedence constraints.
        # Move predecessor tasks that have been removed to ext_predecessors.
        for tinfo in partial_tasks.values():
            # Keep predecessors if they are still in the graph.
            new_predecessors = {
                pred for pred in tinfo.predecessors if pred in partial_tasks
            }
            # Move others to ext_predecessors.
            new_ext_predecessors = {
                pred for pred in tinfo.predecessors if pred not in partial_tasks
            }
            tinfo.predecessors = new_predecessors
            tinfo.ext_predecessors = new_ext_predecessors
            # Clear successors. Will be filled in at the end of this function.
            tinfo.successors.clear()

        # Precedence constraints for same-processor tasks that used to have a
        # precedence chain of other-processor tasks in between them.
        for tid, tinfo in partial_tasks.items():
            for pred in self.double_cross_predecessors(tid):
                if pred not in tinfo.predecessors:
                    tinfo.predecessors.add(pred)

            # Relative deadlines.
            # Keep rel_deadline to pred if pred is still in the graph.
            new_rel_deadlines = {
                pred: dl
                for pred, dl in tinfo.rel_deadlines.items()
                if pred in partial_tasks
            }
            # Move others to ext_predecessors.
            tinfo.ext_rel_deadlines = {
                pred: dl
                for pred, dl in tinfo.rel_deadlines.items()
                if pred not in partial_tasks
            }
            tinfo.rel_deadlines = new_rel_deadlines

        partial_graph = TaskGraph(partial_tasks)
        # Fill in successors by taking opposite of predecessors.
        partial_graph.update_successors()
        return partial_graph


class TaskGraphBuilder:
    @classmethod
    def linear_tasks(cls, tasks: List[QoalaTask]) -> TaskGraph:
        tinfos: List[TaskInfo] = [TaskInfo.only_task(task) for task in tasks]

        for i in range(len(tinfos) - 1):
            t1 = tinfos[i]
            t2 = tinfos[i + 1]
            t2.predecessors.add(t1.task.task_id)

        graph = TaskGraph(tasks={t.task.task_id: t for t in tinfos})
        graph.update_successors()
        return graph

    @classmethod
    def linear_tasks_with_start_times(
        cls, tasks: List[Tuple[QoalaTask, Optional[int]]]
    ) -> TaskGraph:
        tinfos: List[TaskInfo] = []
        for task, start_time in tasks:
            tinfo = TaskInfo.only_task(task)
            tinfo.start_time = start_time
            tinfos.append(tinfo)

        for i in range(len(tinfos) - 1):
            t1 = tinfos[i]
            t2 = tinfos[i + 1]
            t2.predecessors.add(t1.task.task_id)

        graph = TaskGraph(tasks={t.task.task_id: t for t in tinfos})
        graph.update_successors()
        return graph

    @classmethod
    def merge(cls, graphs: List[TaskGraph]) -> TaskGraph:
        merged_tinfos = {}
        for graph in graphs:
            for tid, tinfo in graph.get_tasks().items():
                merged_tinfos[tid] = tinfo

        merged = TaskGraph(merged_tinfos)
        merged.update_successors()
        return merged

    @classmethod
    def merge_linear(cls, graphs: List[TaskGraph]) -> TaskGraph:
        merged_tinfos = {}
        for graph in graphs:
            for tid, tinfo in graph.get_tasks().items():
                merged_tinfos[tid] = deepcopy(tinfo)

        merged = TaskGraph(merged_tinfos)

        for i in range(1, len(graphs)):
            chain1 = graphs[i - 1].linearize()
            chain2 = graphs[i].linearize()
            # Add precedence between last task of graph1 and first task of graph2
            precedence = (chain1[-1], chain2[0])
            merged.add_precedences([precedence])

        merged.update_successors()
        return merged

    @classmethod
    def from_program(
        cls,
        program: QoalaProgram,
        pid: int,
        ehi: Optional[EhiNodeInfo] = None,
        network_ehi: Optional[EhiNetworkInfo] = None,
        first_task_id: int = 0,
        prog_input: Optional[Dict[str, int]] = None,
    ) -> TaskGraph:
        return QoalaGraphFromProgramBuilder(first_task_id).build(
            program, pid, ehi, network_ehi, prog_input
        )


class TaskDurationEstimator:
    @classmethod
    def lr_duration(cls, ehi: EhiNodeInfo, routine: LocalRoutine) -> float:
        duration = 0.0
        # TODO: refactor this
        for instr in routine.subroutine.instructions:
            if (
                type(instr)
                in [
                    core.SetInstruction,
                    core.StoreInstruction,
                    core.LoadInstruction,
                    core.LeaInstruction,
                ]
                or isinstance(instr, core.BranchBinaryInstruction)
                or isinstance(instr, core.BranchUnaryInstruction)
                or isinstance(instr, core.JmpInstruction)
                or isinstance(instr, core.ClassicalOpInstruction)
                or isinstance(instr, core.ClassicalOpModInstruction)
            ):
                duration += ehi.latencies.qnos_instr_time
            else:
                max_duration = -1.0
                # TODO: gate duration depends on which qubit!!
                # currently we always take the worst case scenario but this is not ideal
                for i in ehi.single_gate_infos.keys():
                    if info := ehi.find_single_gate(i, type(instr)):
                        max_duration = max(max_duration, info.duration)

                for multi in ehi.multi_gate_infos.keys():
                    if info := ehi.find_multi_gate(multi.qubit_ids, type(instr)):
                        max_duration = max(max_duration, info.duration)

                if ehi.all_qubit_gate_infos is not None:
                    for gate_info in ehi.all_qubit_gate_infos:
                        max_duration = max(max_duration, gate_info.duration)

                if max_duration != -1:
                    duration += max_duration
                else:
                    raise RuntimeError(
                        f"Gate {type(instr)} not found in EHI. Cannot calculate duration of containing block."
                    )
        return duration


class TaskGraphFromBlockBuilder:
    def __init__(self):
        self._task_id_counter: int = 0

    def unique_id(self) -> int:
        task_id = self._task_id_counter
        self._task_id_counter += 1
        return task_id

    def build(
        self,
        program_instance: ProgramInstance,
        block_index: int,
        network_ehi: Optional[EhiNetworkInfo] = None,
    ) -> TaskGraph:
        graph = TaskGraph()
        block = program_instance.program.blocks[block_index]
        pid = program_instance.pid
        local_routines = program_instance.program.local_routines
        request_routines = program_instance.program.request_routines
        ehi = program_instance.unit_module.info
        prog_input = program_instance.inputs.values

        if block.typ == BasicBlockType.CL:
            if ehi is not None:
                duration = ehi.latencies.host_instr_time * len(block.instructions)
            else:
                duration = None
            task_id = self.unique_id()
            graph.add_tasks([HostLocalTask(task_id, pid, block.name, duration)])

            if block.deadlines is not None:
                graph.get_tinfo(task_id).deadline = 0
        elif block.typ == BasicBlockType.CC:
            assert len(block.instructions) == 1
            instr = block.instructions[0]
            assert isinstance(instr, ReceiveCMsgOp)

            if ehi is not None:
                duration = ehi.latencies.host_peer_latency
            else:
                duration = None
            task_id = self.unique_id()
            graph.add_tasks([HostEventTask(task_id, pid, block.name, duration)])
            if block.deadlines is not None:
                # TODO: fix this hack
                graph.get_tinfo(task_id).deadline = 0
        elif block.typ == BasicBlockType.QL:
            assert len(block.instructions) == 1
            instr = block.instructions[0]
            assert isinstance(instr, RunSubroutineOp)
            if ehi is not None:
                local_routine = local_routines[instr.subroutine]
                lr_duration = TaskDurationEstimator.lr_duration(ehi, local_routine)
                pre_duration = ehi.latencies.host_instr_time
                post_duration = ehi.latencies.host_instr_time
            else:
                lr_duration = None
                pre_duration = None
                post_duration = None

            precall_id = self.unique_id()
            # Use a unique "pointer" or identifier which is used at runtime to point
            # to shared data. The PreCallTask will store the lrcall object
            # to this location, such that the LR- and postcall task can
            # access this object using the shared pointer.
            shared_ptr = precall_id  # just use this task id so we know it's unique
            precall_task = PreCallTask(
                precall_id, pid, block.name, shared_ptr, pre_duration
            )
            graph.add_tasks([precall_task])

            lr_id = self.unique_id()
            qputask = LocalRoutineTask(lr_id, pid, block.name, shared_ptr, lr_duration)
            graph.add_tasks([qputask])

            postcall_id = self.unique_id()
            postcall_task = PostCallTask(
                postcall_id, pid, block.name, shared_ptr, post_duration
            )
            graph.add_tasks([postcall_task])

            # LR task should come after precall task
            graph.get_tinfo(lr_id).predecessors.add(precall_id)
            # postcall task should come after LR task
            graph.get_tinfo(postcall_id).predecessors.add(lr_id)

            if block.deadlines is not None:
                # TODO: fix this hack
                graph.get_tinfo(precall_id).deadline = 0
                graph.get_tinfo(lr_id).deadline = 0
                graph.get_tinfo(postcall_id).deadline = 0
        elif block.typ == BasicBlockType.QC:
            assert len(block.instructions) == 1
            instr = block.instructions[0]
            assert isinstance(instr, RunRequestOp)
            req_routine = request_routines[instr.req_routine]
            callback = req_routine.callback

            if ehi is not None:
                # TODO: make more accurate!
                pre_duration = ehi.latencies.host_instr_time
                post_duration = ehi.latencies.host_instr_time
                cb_duration = ehi.latencies.qnos_instr_time
            else:
                pre_duration = None
                post_duration = None
                cb_duration = None

            if network_ehi is not None:
                pair_duration = list(network_ehi.links.values())[0].duration
                num_pairs = req_routine.request.num_pairs
                if isinstance(num_pairs, Template):
                    assert prog_input is not None
                    num_pairs = prog_input[num_pairs.name]
                multi_duration = pair_duration * num_pairs
            else:
                pair_duration = None
                multi_duration = None

            precall_id = self.unique_id()
            # Use a unique "pointer" or identifier which is used at runtime to point
            # to shared data. The PreCallTask will store the lrcall or rrcall object
            # to this location, such that the pair- callback- and postcall tasks can
            # access this object using the shared pointer.
            shared_ptr = precall_id  # just use this task id so we know it's unique
            precall_task = PreCallTask(
                precall_id, pid, block.name, shared_ptr, pre_duration
            )
            graph.add_tasks([precall_task])

            postcall_id = self.unique_id()
            postcall_task = PostCallTask(
                postcall_id, pid, block.name, shared_ptr, post_duration
            )
            graph.add_tasks([postcall_task])

            if block.deadlines is not None:
                # TODO: fix this hack
                graph.get_tinfo(precall_id).deadline = 0
                graph.get_tinfo(postcall_id).deadline = 0

            if req_routine.callback_type == CallbackType.WAIT_ALL:
                rr_id = self.unique_id()
                rr_task = MultiPairTask(rr_id, pid, shared_ptr, multi_duration)
                graph.add_tasks([rr_task])
                # RR task should come after precall task
                graph.get_tinfo(rr_id).predecessors.add(precall_id)

                if callback is not None:
                    cb_id = self.unique_id()
                    cb_task = MultiPairCallbackTask(
                        cb_id, pid, callback, shared_ptr, cb_duration
                    )
                    graph.add_tasks([cb_task])
                    if block.deadlines is not None:
                        # TODO: fix this hack
                        graph.get_tinfo(cb_id).deadline = 0
                    # callback task should come after RR task
                    graph.get_tinfo(cb_id).predecessors.add(rr_id)
                    # postcall task should come after callback task
                    graph.get_tinfo(postcall_id).predecessors.add(cb_id)
                else:  # no callback
                    # postcall task should come after RR task
                    graph.get_tinfo(postcall_id).predecessors.add(rr_id)

            else:
                assert req_routine.callback_type == CallbackType.SEQUENTIAL

                num_pairs = req_routine.request.num_pairs
                if isinstance(num_pairs, Template):
                    assert prog_input is not None
                    num_pairs = prog_input[num_pairs.name]

                for i in range(num_pairs):
                    rr_pair_id = self.unique_id()
                    rr_pair_task = SinglePairTask(
                        rr_pair_id, pid, i, shared_ptr, pair_duration
                    )
                    graph.add_tasks([rr_pair_task])
                    if block.deadlines is not None:
                        # TODO: fix this hack
                        graph.get_tinfo(rr_pair_id).deadline = 0
                    # RR pair task should come after precall task.
                    # Note: the RR pair tasks do not have precedence
                    # constraints among each other.
                    graph.get_tinfo(rr_pair_id).predecessors.add(precall_id)
                    if callback is not None:
                        pair_cb_id = self.unique_id()
                        pair_cb_task = SinglePairCallbackTask(
                            pair_cb_id, pid, callback, i, shared_ptr, cb_duration
                        )
                        graph.add_tasks([pair_cb_task])
                        if block.deadlines is not None:
                            # TODO: fix this hack
                            graph.get_tinfo(pair_cb_id).deadline = 0
                        # Callback task for pair should come after corresponding
                        # RR pair task. Note: the pair callback tasks do not have
                        # precedence constraints among each other.
                        graph.get_tinfo(pair_cb_id).predecessors.add(rr_pair_id)
                        # postcall task should come after callback task
                        graph.get_tinfo(postcall_id).predecessors.add(pair_cb_id)
                    else:  # no callback
                        # postcall task should come after RR task
                        graph.get_tinfo(postcall_id).predecessors.add(rr_pair_id)

        return graph


class QoalaGraphFromProgramBuilder:
    def __init__(self, first_task_id: int = 0) -> None:
        self._first_task_id = first_task_id
        self._task_id_counter = first_task_id
        self._graph = TaskGraph()
        self._block_to_task_map: Dict[str, int] = {}  # blk name -> task ID

    def unique_id(self) -> int:
        id = self._task_id_counter
        self._task_id_counter += 1
        return id

    def build(
        self,
        program: QoalaProgram,
        pid: int,
        ehi: Optional[EhiNodeInfo] = None,
        network_ehi: Optional[EhiNetworkInfo] = None,
        prog_input: Optional[Dict[str, int]] = None,
    ) -> TaskGraph:
        prev_block_task_id: Optional[int] = None
        for block in program.blocks:
            if block.typ == BasicBlockType.CL:
                if ehi is not None:
                    duration = ehi.latencies.host_instr_time * len(block.instructions)
                else:
                    duration = None
                task_id = self.unique_id()
                self._graph.add_tasks(
                    [HostLocalTask(task_id, pid, block.name, duration)]
                )
                self._block_to_task_map[block.name] = task_id
                # Task for this block should come after task for previous block
                # (Assuming linear program!)
                if prev_block_task_id is not None:
                    self._graph.get_tinfo(task_id).predecessors.add(prev_block_task_id)
                if block.deadlines is not None:
                    for blk, dl in block.deadlines.items():
                        other_task = self._block_to_task_map[blk]
                        self._graph.get_tinfo(task_id).rel_deadlines[other_task] = dl
                prev_block_task_id = task_id
            elif block.typ == BasicBlockType.CC:
                assert len(block.instructions) == 1
                instr = block.instructions[0]
                assert isinstance(instr, ReceiveCMsgOp)
                if ehi is not None:
                    duration = ehi.latencies.host_peer_latency
                else:
                    duration = None
                task_id = self.unique_id()
                self._graph.add_tasks(
                    [HostEventTask(task_id, pid, block.name, duration)]
                )
                self._block_to_task_map[block.name] = task_id
                # Task for this block should come after task for previous block
                # (Assuming linear program!)
                if prev_block_task_id is not None:
                    self._graph.get_tinfo(task_id).predecessors.add(prev_block_task_id)
                prev_block_task_id = task_id
            elif block.typ == BasicBlockType.QL:
                assert len(block.instructions) == 1
                instr = block.instructions[0]
                assert isinstance(instr, RunSubroutineOp)
                if ehi is not None:
                    local_routine = program.local_routines[instr.subroutine]
                    lr_duration = TaskDurationEstimator.lr_duration(ehi, local_routine)
                    pre_duration = ehi.latencies.host_instr_time
                    post_duration = ehi.latencies.host_instr_time
                else:
                    lr_duration = None
                    pre_duration = None
                    post_duration = None

                deadlines: Dict[int, int] = {}  # other task ID -> relative deadline
                if block.deadlines is not None:
                    for blk, dl in block.deadlines.items():
                        other_task = self._block_to_task_map[blk]
                        deadlines[other_task] = dl

                precall_id = self.unique_id()
                # Use a unique "pointer" or identifier which is used at runtime to point
                # to shared data. The PreCallTask will store the lrcall object
                # to this location, such that the LR- and postcall task can
                # access this object using the shared pointer.
                shared_ptr = precall_id  # just use this task id so we know it's unique
                precall_task = PreCallTask(
                    precall_id, pid, block.name, shared_ptr, pre_duration
                )
                self._graph.add_tasks([precall_task])
                for other_task, dl in deadlines.items():
                    # TODO: fix this hack
                    self._graph.get_tinfo(precall_id).rel_deadlines[other_task] = dl

                lr_id = self.unique_id()
                qputask = LocalRoutineTask(
                    lr_id, pid, block.name, shared_ptr, lr_duration
                )
                self._graph.add_tasks([qputask])

                postcall_id = self.unique_id()
                postcall_task = PostCallTask(
                    postcall_id, pid, block.name, shared_ptr, post_duration
                )
                self._graph.add_tasks([postcall_task])
                self._block_to_task_map[block.name] = postcall_id

                # LR task should come after precall task
                self._graph.get_tinfo(lr_id).predecessors.add(precall_id)
                # postcall task should come after LR task
                self._graph.get_tinfo(postcall_id).predecessors.add(lr_id)

                # Tasks for this block should come after task for previous block
                # (Assuming linear program!)
                if prev_block_task_id is not None:
                    # First task for this block is precall task.
                    self._graph.get_tinfo(precall_id).predecessors.add(
                        prev_block_task_id
                    )
                # Last task for this block is postcall task.
                prev_block_task_id = postcall_id
            elif block.typ == BasicBlockType.QC:
                precall_id, postcall_id = self._build_from_qc_task_routine_split(
                    program, block, pid, ehi, network_ehi, prog_input
                )
                # Tasks for this block should come after task for previous block
                # (Assuming linear program!)
                if prev_block_task_id is not None:
                    # First task for QC block is precall task.
                    self._graph.get_tinfo(precall_id).predecessors.add(
                        prev_block_task_id
                    )
                # Last task for QC block is postcall task.
                prev_block_task_id = postcall_id  # (not precall_id !)

        self._graph.update_successors()
        return self._graph

    def _build_from_qc_task_routine_split(
        self,
        program: QoalaProgram,
        block: BasicBlock,
        pid: int,
        ehi: Optional[EhiNodeInfo] = None,
        network_ehi: Optional[EhiNetworkInfo] = None,
        prog_input: Optional[Dict[str, int]] = None,
    ) -> Tuple[int, int]:
        """Returns (precall_id, post_call_id)"""
        assert len(block.instructions) == 1
        instr = block.instructions[0]
        assert isinstance(instr, RunRequestOp)
        req_routine = program.request_routines[instr.req_routine]
        callback_name = req_routine.callback

        if ehi is not None:
            # TODO: make more accurate!
            pre_duration = ehi.latencies.host_instr_time
            post_duration = ehi.latencies.host_instr_time
            if callback_name is not None:
                callback = program.local_routines[callback_name]
                cb_duration = TaskDurationEstimator.lr_duration(ehi, callback)
        else:
            pre_duration = None
            post_duration = None
            cb_duration = None

        if network_ehi is not None:
            pair_duration = list(network_ehi.links.values())[0].duration
            num_pairs = req_routine.request.num_pairs
            if isinstance(num_pairs, Template):
                assert prog_input is not None
                num_pairs = prog_input[num_pairs.name]
            multi_duration = pair_duration * num_pairs
        else:
            pair_duration = None
            multi_duration = None

        precall_id = self.unique_id()
        # Use a unique "pointer" or identifier which is used at runtime to point
        # to shared data. The PreCallTask will store the lrcall or rrcall object
        # to this location, such that the pair- callback- and postcall tasks can
        # access this object using the shared pointer.
        shared_ptr = precall_id  # just use this task id so we know it's unique
        precall_task = PreCallTask(
            precall_id, pid, block.name, shared_ptr, pre_duration
        )
        self._graph.add_tasks([precall_task])

        postcall_id = self.unique_id()
        postcall_task = PostCallTask(
            postcall_id, pid, block.name, shared_ptr, post_duration
        )
        self._graph.add_tasks([postcall_task])
        self._block_to_task_map[block.name] = postcall_id

        if req_routine.callback_type == CallbackType.WAIT_ALL:
            rr_id = self.unique_id()
            rr_task = MultiPairTask(rr_id, pid, shared_ptr, multi_duration)
            self._graph.add_tasks([rr_task])
            # RR task should come after precall task
            self._graph.get_tinfo(rr_id).predecessors.add(precall_id)

            if callback_name is not None:
                cb_id = self.unique_id()
                cb_task = MultiPairCallbackTask(
                    cb_id, pid, callback_name, shared_ptr, cb_duration
                )
                self._graph.add_tasks([cb_task])
                # callback task should come after RR task
                self._graph.get_tinfo(cb_id).predecessors.add(rr_id)
                # postcall task should come after callback task
                self._graph.get_tinfo(postcall_id).predecessors.add(cb_id)
            else:  # no callback
                # postcall task should come after RR task
                self._graph.get_tinfo(postcall_id).predecessors.add(rr_id)

        else:
            assert req_routine.callback_type == CallbackType.SEQUENTIAL

            num_pairs = req_routine.request.num_pairs
            if isinstance(num_pairs, Template):
                assert prog_input is not None
                num_pairs = prog_input[num_pairs.name]

            for i in range(num_pairs):
                rr_pair_id = self.unique_id()
                rr_pair_task = SinglePairTask(
                    rr_pair_id, pid, i, shared_ptr, pair_duration
                )
                self._graph.add_tasks([rr_pair_task])
                # RR pair task should come after precall task.
                # Note: the RR pair tasks do not have precedence
                # constraints among each other.
                self._graph.get_tinfo(rr_pair_id).predecessors.add(precall_id)
                if callback_name is not None:
                    pair_cb_id = self.unique_id()
                    pair_cb_task = SinglePairCallbackTask(
                        pair_cb_id, pid, callback_name, i, shared_ptr, cb_duration
                    )
                    self._graph.add_tasks([pair_cb_task])
                    # Callback task for pair should come after corresponding
                    # RR pair task. Note: the pair callback tasks do not have
                    # precedence constraints among each other.
                    self._graph.get_tinfo(pair_cb_id).predecessors.add(rr_pair_id)
                    # postcall task should come after callback task
                    self._graph.get_tinfo(postcall_id).predecessors.add(pair_cb_id)
                else:  # no callback
                    # postcall task should come after RR task
                    self._graph.get_tinfo(postcall_id).predecessors.add(rr_pair_id)

        return precall_id, postcall_id



from __future__ import annotations

import os

from netqasm.lang.instr.flavour import NVFlavour

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    NvParams,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import ProgramInput
from qoala.util.runner import run_single_node_app


def get_config() -> ProcNodeConfig:
    params = NvParams()
    topology = TopologyConfig.from_nv_params(num_qubits=5, params=params)
    return ProcNodeConfig(
        node_name="alice",
        node_id=0,
        topology=topology,
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("NvNtf"),
    )


def load_program(name: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), name)
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text, flavour=NVFlavour()).parse()

    return program


def run_program(num_iterations: int, node_cfg: ProcNodeConfig):
    network_cfg = ProcNodeNetworkConfig(nodes=[node_cfg], links=[])
    program = load_program("move_qubit_nv.iqoala")

    app_results = run_single_node_app(
        num_iterations=num_iterations,
        program_name="alice",
        program=program,
        program_input=ProgramInput.empty(),
        network_cfg=network_cfg,
        linear=True,
    )

    all_results = app_results.batch_results["alice"].results
    outcomes = [result.values["m"] for result in all_results]
    assert all(outcome == 1 for outcome in outcomes)


def test_move_qubit():
    node_cfg = get_config()
    run_program(10, node_cfg)


if __name__ == "__main__":
    test_move_qubit()



import pytest
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.models.qerrormodels import DepolarNoiseModel, T1T2NoiseModel
from netsquid.components.qprocessor import MissingInstructionError, QuantumProcessor

from qoala.lang.ehi import EhiLinkInfo, EhiNetworkInfo
from qoala.runtime.config import (
    LatenciesConfig,
    LinkBetweenNodesConfig,
    LinkConfig,
    NtfConfig,
    NvParams,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.instructions import INSTR_MEASURE_ALL, INSTR_ROT_X_ALL
from qoala.runtime.lhi import (
    LhiGateInfo,
    LhiLatencies,
    LhiNetworkInfo,
    LhiProcNodeInfo,
    LhiQubitInfo,
    LhiTopology,
    LhiTopologyBuilder,
)
from qoala.runtime.ntf import GenericNtf
from qoala.sim.build import (
    build_network_from_config,
    build_network_from_lhi,
    build_procnode_from_config,
    build_procnode_from_lhi,
    build_qprocessor_from_topology,
)


def uniform_topology(num_qubits: int) -> LhiTopology:
    qubit_info = LhiQubitInfo(
        is_communication=True,
        error_model=T1T2NoiseModel,
        error_model_kwargs={"T1": 1e6, "T2": 1e6},
    )
    single_gate_infos = [
        LhiGateInfo(
            instruction=instr,
            duration=5e3,
            error_model=DepolarNoiseModel,
            error_model_kwargs={
                "depolar_rate": 0.2,
                "time_independent": True,
            },
        )
        for instr in [INSTR_X, INSTR_Y, INSTR_Z]
    ]
    two_gate_infos = [
        LhiGateInfo(
            instruction=INSTR_CNOT,
            duration=2e4,
            error_model=DepolarNoiseModel,
            error_model_kwargs={
                "depolar_rate": 0.2,
                "time_independent": True,
            },
        )
    ]

    return LhiTopologyBuilder.fully_uniform(
        num_qubits=num_qubits,
        qubit_info=qubit_info,
        single_gate_infos=single_gate_infos,
        two_gate_infos=two_gate_infos,
    )


def test_build_from_topology():
    num_qubits = 3
    topology = uniform_topology(num_qubits)
    proc: QuantumProcessor = build_qprocessor_from_topology("proc", topology)
    assert proc.num_positions == num_qubits

    assert proc.name == "proc"

    for i in range(num_qubits):
        assert (
            proc.get_instruction_duration(INSTR_X, [i])
            == topology.find_single_gate(i, INSTR_X).duration
        )
        with pytest.raises(MissingInstructionError):
            proc.get_instruction_duration(INSTR_ROT_X, [i])

    assert (
        proc.get_instruction_duration(INSTR_CNOT, [0, 1])
        == topology.find_multi_gate([0, 1], INSTR_CNOT).duration
    )


def test_build_from_topology_with_all_qubit_gates():
    num_qubits = 3
    topology = LhiTopologyBuilder.perfect_uniform(
        num_qubits=num_qubits,
        single_instructions=[INSTR_INIT, INSTR_MEASURE, INSTR_ROT_Z],
        single_duration=5e3,
        two_instructions=[],
        two_duration=100e3,
        all_qubit_instructions=[INSTR_ROT_X_ALL, INSTR_MEASURE_ALL],
        all_qubit_duration=50e3,
    )
    proc: QuantumProcessor = build_qprocessor_from_topology("proc", topology)
    assert proc.num_positions == num_qubits

    for i in range(num_qubits):
        assert (
            proc.get_instruction_duration(INSTR_ROT_Z, [i])
            == topology.find_single_gate(i, INSTR_ROT_Z).duration
        )
        with pytest.raises(MissingInstructionError):
            proc.get_instruction_duration(INSTR_X, [i])

    assert proc.get_instruction_duration(INSTR_ROT_X_ALL, [0, 1, 2]) == 50e3
    assert proc.get_instruction_duration(INSTR_MEASURE_ALL, [0, 1, 2]) == 50e3


def test_build_perfect_topology():
    num_qubits = 3
    topology = LhiTopologyBuilder.perfect_uniform(
        num_qubits=num_qubits,
        single_instructions=[INSTR_X, INSTR_Y],
        single_duration=5e3,
        two_instructions=[INSTR_CNOT],
        two_duration=100e3,
    )
    proc: QuantumProcessor = build_qprocessor_from_topology("proc", topology)
    assert proc.num_positions == num_qubits

    assert proc.name == "proc"

    for i in range(num_qubits):
        assert (
            proc.get_instruction_duration(INSTR_X, [i])
            == topology.find_single_gate(i, INSTR_X).duration
        )
        assert proc.get_instruction_duration(INSTR_X, [i]) == 5e3

        with pytest.raises(MissingInstructionError):
            proc.get_instruction_duration(INSTR_ROT_X, [i])

    assert (
        proc.get_instruction_duration(INSTR_CNOT, [0, 1])
        == topology.find_multi_gate([0, 1], INSTR_CNOT).duration
    )
    assert proc.get_instruction_duration(INSTR_CNOT, [0, 1]) == 100e3


def test_build_nv_perfect():
    num_qubits = 2
    parms = NvParams()
    parms.comm_init_duration = 500
    parms.comm_meas_duration = 1200
    parms.comm_gate_duration = 100
    parms.mem_init_duration = 35000
    parms.mem_meas_duration = 80000
    parms.mem_gate_duration = 25000
    parms.two_gate_duration = 100_000
    cfg = TopologyConfig.from_nv_params(num_qubits, parms)
    topology = LhiTopologyBuilder.from_config(cfg)
    proc: QuantumProcessor = build_qprocessor_from_topology("alice", topology)
    assert proc.num_positions == num_qubits

    assert proc.get_instruction_duration(INSTR_INIT, [0]) == parms.comm_init_duration
    assert proc.get_instruction_duration(INSTR_MEASURE, [0]) == parms.comm_meas_duration
    assert proc.get_instruction_duration(INSTR_ROT_X, [0]) == parms.comm_gate_duration

    for i in range(1, num_qubits):
        assert proc.get_instruction_duration(INSTR_INIT, [i]) == parms.mem_init_duration
        assert (
            proc.get_instruction_duration(INSTR_ROT_X, [i]) == parms.mem_gate_duration
        )
        assert (
            proc.get_instruction_duration(INSTR_MEASURE, [i]) == parms.mem_meas_duration
        )

    with pytest.raises(MissingInstructionError):
        proc.get_instruction_duration(INSTR_CNOT, [0, 1])
        proc.get_instruction_duration(INSTR_CXDIR, [1, 0])

    assert proc.get_instruction_duration(INSTR_CXDIR, [0, 1]) == parms.two_gate_duration


def test_build_procnode_from_config():
    top_cfg = TopologyConfig.perfect_config_uniform_default_params(num_qubits=2)
    latencies = LatenciesConfig(
        host_instr_time=17,
        qnos_instr_time=20,
        host_peer_latency=5,
    )
    cfg = ProcNodeConfig(
        node_name="the_node",
        node_id=42,
        topology=top_cfg,
        latencies=latencies,
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )
    nodes = {42: "the_node", 43: "other_node"}
    network_ehi = EhiNetworkInfo.perfect_fully_connected(nodes, duration=1000)

    procnode = build_procnode_from_config(cfg, network_ehi)

    assert procnode.node.name == "the_node"
    procnode.host_comp.peer_in_port("other_node")  # should not raise error
    procnode.netstack_comp.peer_in_port("other_node")  # should not raise error
    assert procnode.qnos.processor._latencies.qnos_instr_time == 20
    assert procnode.host.processor._latencies.host_instr_time == 17
    assert procnode.host.processor._latencies.host_peer_latency == 5

    expected_topology = LhiTopologyBuilder.from_config(top_cfg)
    expected_qprocessor = build_qprocessor_from_topology("the_node", expected_topology)
    actual_qprocessor = procnode.qdevice.qprocessor

    assert expected_qprocessor.name == actual_qprocessor.name
    assert expected_qprocessor.num_positions == actual_qprocessor.num_positions

    assert expected_qprocessor.get_instruction_duration(
        INSTR_X, [0]
    ) == actual_qprocessor.get_instruction_duration(INSTR_X, [0])

    assert expected_topology == procnode.qdevice.topology

    assert procnode.network_ehi.get_link(42, 43) == EhiLinkInfo(1000, 1.0)


def test_build_network_from_config():
    top_cfg = TopologyConfig.perfect_config_uniform_default_params(num_qubits=2)
    cfg_alice = ProcNodeConfig(
        node_name="alice",
        node_id=42,
        topology=top_cfg,
        latencies=LatenciesConfig(),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )
    cfg_bob = ProcNodeConfig(
        node_name="bob",
        node_id=43,
        topology=top_cfg,
        latencies=LatenciesConfig(),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )

    link_cfg = LinkConfig.perfect_config(state_delay=1000)
    link_ab = LinkBetweenNodesConfig(node_id1=42, node_id2=43, link_config=link_cfg)
    cfg = ProcNodeNetworkConfig(nodes=[cfg_alice, cfg_bob], links=[link_ab])
    network = build_network_from_config(cfg)

    assert len(network.nodes) == 2
    assert "alice" in network.nodes
    assert "bob" in network.nodes
    assert network.entdist is not None

    alice = network.nodes["alice"]
    bob = network.nodes["bob"]
    entdist = network.entdist

    assert entdist.get_sampler(42, 43).delay == 1000

    # NOTE: alice.host_comp.peer_in_port("bob") does not have a 'connected_port'.
    # Rather, messages are forwarded from alice.node.host_peer_in_port("bob").

    alice_host_in = alice.node.host_peer_in_port("bob")
    alice_host_out = alice.node.host_peer_out_port("bob")
    bob_host_in = bob.node.host_peer_in_port("alice")
    bob_host_out = bob.node.host_peer_out_port("alice")

    alice_host_in_chan = alice_host_in.connected_port.component
    alice_host_in_remote = alice_host_in_chan.ports["send"].connected_port
    assert alice_host_in_remote == bob_host_out

    alice_host_out_chan = alice_host_out.connected_port.component
    alice_host_out_remote = alice_host_out_chan.ports["recv"].connected_port
    assert alice_host_out_remote == bob_host_in

    alice_ent_in = alice.node.entdist_in_port
    alice_ent_out = alice.node.entdist_out_port
    bob_ent_in = bob.node.entdist_in_port
    bob_ent_out = bob.node.entdist_out_port

    alice_ent_in_chan = alice_ent_in.connected_port.component
    alice_ent_in_remote = alice_ent_in_chan.ports["send"].connected_port
    assert alice_ent_in_remote == entdist.comp.node_out_port("alice")

    alice_ent_out_chan = alice_ent_out.connected_port.component
    alice_ent_out_remote = alice_ent_out_chan.ports["recv"].connected_port
    assert alice_ent_out_remote == entdist.comp.node_in_port("alice")

    bob_ent_in_chan = bob_ent_in.connected_port.component
    bob_ent_in_remote = bob_ent_in_chan.ports["send"].connected_port
    assert bob_ent_in_remote == entdist.comp.node_out_port("bob")

    bob_ent_out_chan = bob_ent_out.connected_port.component
    bob_ent_out_remote = bob_ent_out_chan.ports["recv"].connected_port
    assert bob_ent_out_remote == entdist.comp.node_in_port("bob")


def test_build_network_perfect_links():
    top_cfg = TopologyConfig.perfect_config_uniform_default_params(num_qubits=2)
    cfg_alice = ProcNodeConfig(
        node_name="alice",
        node_id=42,
        topology=top_cfg,
        latencies=LatenciesConfig(),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )
    cfg_bob = ProcNodeConfig(
        node_name="bob",
        node_id=43,
        topology=top_cfg,
        latencies=LatenciesConfig(),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )
    cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[cfg_alice, cfg_bob], link_duration=500
    )
    network = build_network_from_config(cfg)

    assert len(network.nodes) == 2
    assert "alice" in network.nodes
    assert "bob" in network.nodes
    assert network.entdist is not None

    entdist = network.entdist
    assert entdist.get_sampler(42, 43).delay == 500


def test_build_network_from_lhi():
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits=3)
    latencies = LhiLatencies(
        host_instr_time=500,
        qnos_instr_time=1000,
        host_peer_latency=20_000,
    )
    alice_lhi = LhiProcNodeInfo(
        name="alice",
        id=42,
        topology=topology,
        latencies=latencies,
    )
    bob_lhi = LhiProcNodeInfo(
        name="bob",
        id=43,
        topology=topology,
        latencies=latencies,
    )
    nodes = {42: "alice", 43: "bob"}

    network_lhi = LhiNetworkInfo.perfect_fully_connected(nodes, 100_000)
    ntfs = [GenericNtf(), GenericNtf()]
    network = build_network_from_lhi([alice_lhi, bob_lhi], ntfs, network_lhi)

    assert len(network.nodes) == 2
    assert "alice" in network.nodes
    assert "bob" in network.nodes
    assert network.entdist is not None

    alice = network.nodes["alice"]
    entdist = network.entdist

    assert entdist.get_sampler(42, 43).delay == 100_000

    assert alice.local_ehi.latencies.host_instr_time == 500
    assert alice.local_ehi.latencies.qnos_instr_time == 1000
    assert alice.local_ehi.latencies.host_peer_latency == 20_000


def test_build_procnode_from_lhi():
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits=3)
    latencies = LhiLatencies(
        host_instr_time=500,
        qnos_instr_time=1000,
        host_peer_latency=20_000,
    )
    nodes = {42: "alice", 43: "bob"}

    network_lhi = LhiNetworkInfo.perfect_fully_connected(nodes, 100_000)

    alice_procnode = build_procnode_from_lhi(
        name="alice",
        id=42,
        topology=topology,
        latencies=latencies,
        network_lhi=network_lhi,
        ntf=GenericNtf(),
    )

    assert alice_procnode.qdevice.qprocessor.name == "alice_processor"
    assert alice_procnode.qdevice.qprocessor.num_positions == 3

    assert alice_procnode.network_ehi.get_link(42, 43).duration == 100_000

    assert alice_procnode.local_ehi.latencies.host_instr_time == 500
    assert alice_procnode.local_ehi.latencies.qnos_instr_time == 1000
    assert alice_procnode.local_ehi.latencies.host_peer_latency == 20_000

    assert alice_procnode.node.netstack_peer_in_port("bob") is not None
    assert alice_procnode.node.netstack_peer_out_port("bob") is not None


if __name__ == "__main__":
    test_build_from_topology()
    test_build_perfect_topology()
    test_build_nv_perfect()
    test_build_procnode_from_config()
    test_build_network_from_config()
    test_build_network_perfect_links()
    test_build_network_from_lhi()
    test_build_procnode_from_lhi()



import matplotlib
import matplotlib.pyplot as plt
import networkx as nx

from qoala.runtime.task import TaskGraph


class TaskGraphWriter:
    def __init__(self, task_graph: TaskGraph) -> None:
        self._tg = task_graph

        G = nx.DiGraph()

        for task_id, tinfo in task_graph.get_tasks().items():
            typ = tinfo.task.__class__.__name__
            G.add_node(task_id, typ=typ)

        for task_id, tinfo in task_graph.get_tasks().items():
            for pred in tinfo.predecessors:
                G.add_edge(pred, task_id)
            # for succ in tinfo.successors:
            #     G.add_edge(task_id, succ)

        self._nx_graph = G

    def draw(self, path: str) -> None:
        matplotlib.use("Agg")
        f = plt.figure()
        pos = nx.planar_layout(self._nx_graph)
        labels = nx.get_node_attributes(self._nx_graph, "typ")
        nx.draw(
            self._nx_graph,
            pos,
            ax=f.add_subplot(111),
            with_labels=True,
            labels=labels,
            font_size=8,
        )
        f.savefig(path)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, List, Optional, Union

from netqasm.lang.operand import Template

IqoalaValue = Union[int, Template, str]


class HostLanguageSyntaxError(Exception):
    pass


class IqoalaInstructionType(Enum):
    CC = 0
    CL = auto()
    QC = auto()
    QL = auto()


class IqoalaVar:
    pass


@dataclass(frozen=True)
class IqoalaSingleton(IqoalaVar):
    name: str

    def __str__(self):
        return self.name


@dataclass(frozen=True)
class IqoalaTuple(IqoalaVar):
    values: List[str]

    def __str__(self) -> str:
        return f"tuple<{','.join(v for v in self.values)}>"


@dataclass(frozen=True)
class IqoalaVector(IqoalaVar):
    name: str
    size: IqoalaValue

    def __str__(self) -> str:
        return f"{self.name}<{self.size}>"


@dataclass(frozen=True)
class IqoalaVectorElement(IqoalaVar):
    name: str
    index: int


class ClassicalIqoalaOp:
    OP_NAME: str = None  # type: ignore
    TYP: IqoalaInstructionType = None  # type: ignore

    def __init__(
        self,
        arguments: Optional[List[IqoalaVar]] = None,
        results: Optional[IqoalaVar] = None,
        attributes: Optional[List[IqoalaValue]] = None,
    ) -> None:
        # TODO: support list of strs and tuples
        # currently not needed and confuses mypy
        self._arguments: List[IqoalaVar]
        self._results: IqoalaVar
        self._attributes: List[IqoalaValue]

        if arguments is None:
            self._arguments = []  # type: ignore
        else:
            self._arguments = arguments

        self._results = results  # type: ignore

        if attributes is None:
            self._attributes = []
        else:
            self._attributes = attributes

    def __str__(self) -> str:
        if self.results is None:
            results = ""
        else:
            results = str(self.results)
        # not to write  for the empty tuple
        if self.arguments == [IqoalaTuple([])]:
            args = ""
        else:
            args = ", ".join(str(a) for a in self.arguments)
        attrs = ", ".join(str(a) for a in self.attributes)
        s = ""
        if len(results) > 0:
            s += f"{results} = "

        s += f"{self.op_name}({args})"

        if len(attrs) > 0:
            s += f" : {attrs}"
        return s

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ClassicalIqoalaOp):
            return NotImplemented
        return (
            self.results == other.results
            and self.arguments == other.arguments
            and self.attributes == other.attributes
        )

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ) -> ClassicalIqoalaOp:
        raise NotImplementedError

    @property
    def op_name(self) -> str:
        return self.__class__.OP_NAME  # type: ignore

    @property
    def arguments(self) -> List[IqoalaVar]:
        return self._arguments

    @property
    def results(self) -> IqoalaVar:
        return self._results

    @property
    def attributes(self) -> List[IqoalaValue]:
        return self._attributes


class AssignCValueOp(ClassicalIqoalaOp):
    OP_NAME = "assign_cval"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: IqoalaSingleton, value: IqoalaValue) -> None:
        super().__init__(results=result, attributes=[value])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have a result."
            )
        if not isinstance(result, IqoalaSingleton):
            raise HostLanguageSyntaxError
        if len(args) != 0:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 0 arguments but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        return cls(result, attr)


class BusyOp(ClassicalIqoalaOp):
    OP_NAME = "busy"
    TYP = IqoalaInstructionType.CL

    def __init__(self, value: IqoalaValue) -> None:
        super().__init__(attributes=[value])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: Union[List[IqoalaVar]],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have a result."
            )
        if len(args) != 0:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 0 arguments but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        return cls(attr)


class SendCMsgOp(ClassicalIqoalaOp):
    OP_NAME = "send_cmsg"
    TYP = IqoalaInstructionType.CC

    def __init__(
        self,
        csocket: Union[IqoalaSingleton, IqoalaVectorElement],
        value: Union[IqoalaSingleton, IqoalaVectorElement],
    ) -> None:
        # args:
        #   csocket (int): ID of csocket
        #   value (str): name of variable holding the value to send
        super().__init__(arguments=[csocket, value])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: Union[List[IqoalaVar]],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have a result."
            )
        if len(args) != 2:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 2 arguments but got {len(args)}."
            )
        if attr is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have an attribute."
            )
        if (
            not isinstance(args[0], IqoalaSingleton)
            and not isinstance(args[0], IqoalaVectorElement)
        ) or (
            not isinstance(args[1], IqoalaSingleton)
            and not isinstance(args[1], IqoalaVectorElement)
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )
        return cls(args[0], args[1])


class ReceiveCMsgOp(ClassicalIqoalaOp):
    OP_NAME = "recv_cmsg"
    TYP = IqoalaInstructionType.CC

    def __init__(
        self,
        csocket: Union[IqoalaSingleton, IqoalaVectorElement],
        result: IqoalaSingleton,
    ) -> None:
        super().__init__(arguments=[csocket], results=result)

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have a result."
            )
        if not isinstance(result, IqoalaSingleton):
            raise HostLanguageSyntaxError
        if len(args) != 1:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 1 argument but got {len(args)}."
            )
        if attr is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have an attribute."
            )
        if not isinstance(args[0], IqoalaSingleton) and not isinstance(
            args[0], IqoalaVectorElement
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )
        return cls(args[0], result)


class AddCValueOp(ClassicalIqoalaOp):
    OP_NAME = "add_cval_c"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        result: IqoalaSingleton,
        value0: Union[IqoalaSingleton, IqoalaVectorElement],
        value1: Union[IqoalaSingleton, IqoalaVectorElement],
    ) -> None:
        super().__init__(arguments=[value0, value1], results=result)

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have a result."
            )
        if not isinstance(result, IqoalaSingleton):
            raise HostLanguageSyntaxError
        if len(args) != 2:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 2 arguments but got {len(args)}."
            )
        if attr is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have an attribute."
            )
        if (
            not isinstance(args[0], IqoalaSingleton)
            and not isinstance(args[0], IqoalaVectorElement)
        ) or (
            not isinstance(args[1], IqoalaSingleton)
            and not isinstance(args[1], IqoalaVectorElement)
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )
        return cls(result, args[0], args[1])


class MultiplyConstantCValueOp(ClassicalIqoalaOp):
    OP_NAME = "mult_const"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        result: IqoalaSingleton,
        value0: Union[IqoalaSingleton, IqoalaVectorElement],
        const: IqoalaValue,
    ) -> None:
        # result = value0 * const
        super().__init__(arguments=[value0], attributes=[const], results=result)

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have a result."
            )
        if not isinstance(result, IqoalaSingleton):
            raise HostLanguageSyntaxError
        if len(args) != 1:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 1 argument but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if not isinstance(args[0], IqoalaSingleton) and not isinstance(
            args[0], IqoalaVectorElement
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )
        return cls(result, args[0], attr)


class BitConditionalMultiplyConstantCValueOp(ClassicalIqoalaOp):
    OP_NAME = "bcond_mult_const"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        result: IqoalaSingleton,
        value0: Union[IqoalaSingleton, IqoalaVectorElement],
        cond: Union[IqoalaSingleton, IqoalaVectorElement],
        const: int,
    ) -> None:
        # if const == 1:
        #   result = value0 * const
        # else:
        #   result = value0
        super().__init__(arguments=[value0, cond], attributes=[const], results=result)

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have a result."
            )
        if not isinstance(result, IqoalaSingleton):
            raise HostLanguageSyntaxError
        if len(args) != 2:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 2 arguments but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if (
            not isinstance(args[0], IqoalaSingleton)
            and not isinstance(args[0], IqoalaVectorElement)
        ) or (
            not isinstance(args[1], IqoalaSingleton)
            and not isinstance(args[1], IqoalaVectorElement)
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )
        if not isinstance(attr, int):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation attribute must be an integer."
            )
        return cls(result, args[0], args[1], attr)


class RunSubroutineOp(ClassicalIqoalaOp):
    OP_NAME = "run_subroutine"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        result: Optional[Union[IqoalaTuple, IqoalaVector]],
        values: IqoalaTuple,
        subrt: str,
    ) -> None:
        super().__init__(results=result, arguments=[values], attributes=[subrt])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            if not isinstance(result, IqoalaTuple) and not isinstance(
                result, IqoalaVector
            ):
                raise HostLanguageSyntaxError(
                    f"{cls.OP_NAME} operation cannot have a result of type {type(result)}. "
                    f"It must be either IqoalaTuple or IqoalaVector."
                )
        if len(args) == 0:
            args = [IqoalaTuple([])]

        if len(args) != 1:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 1 argument but got {len(args)}."
            )
        if not isinstance(args[0], IqoalaTuple):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation argument must be an IqoalaTuple."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if not isinstance(attr, str):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation attribute must be a string."
            )

        return cls(result, args[0], attr)

    @property
    def subroutine(self) -> str:
        assert isinstance(self.attributes[0], str)
        return self.attributes[0]

    def __str__(self) -> str:
        return super().__str__()


class RunRequestOp(ClassicalIqoalaOp):
    OP_NAME = "run_request"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        result: Optional[Union[IqoalaTuple, IqoalaVector]],
        values: IqoalaTuple,
        routine: str,
    ) -> None:
        super().__init__(results=result, arguments=[values], attributes=[routine])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            if not isinstance(result, IqoalaTuple) and not isinstance(
                result, IqoalaVector
            ):
                raise HostLanguageSyntaxError(
                    f"{cls.OP_NAME} operation cannot have a result of type {type(result)}. "
                    f"It must be either IqoalaTuple or IqoalaVector."
                )
        if len(args) == 0:
            args = [IqoalaTuple([])]

        if len(args) != 1:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 1 argument but got {len(args)}."
            )
        if not isinstance(args[0], IqoalaTuple):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation argument must be an IqoalaTuple."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if not isinstance(attr, str):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation attribute must be a string."
            )

        return cls(result, args[0], attr)

    @property
    def req_routine(self) -> str:
        assert isinstance(self.attributes[0], str)
        return self.attributes[0]

    def __str__(self) -> str:
        return super().__str__()


class ReturnResultOp(ClassicalIqoalaOp):
    OP_NAME = "return_result"
    TYP = IqoalaInstructionType.CL

    def __init__(self, value: Union[IqoalaSingleton, IqoalaVector]) -> None:
        super().__init__(arguments=[value])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have a result."
            )
        if len(args) != 1:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 1 argument but got {len(args)}."
            )
        if attr is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have an attribute."
            )
        if not isinstance(args[0], IqoalaSingleton) and not isinstance(
            args[0], IqoalaVector
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation argument must be a string or vector."
            )

        return cls(args[0])


class JumpOp(ClassicalIqoalaOp):
    OP_NAME = "jump"
    TYP = IqoalaInstructionType.CL

    def __init__(self, block_name: str) -> None:
        super().__init__(attributes=[block_name])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have a result."
            )
        if len(args) != 0:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 0 arguments but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if not isinstance(attr, str):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation attribute must be a string."
            )
        return cls(attr)


class BranchIfEqualOp(ClassicalIqoalaOp):
    OP_NAME = "beq"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        value0: Union[IqoalaSingleton, IqoalaVectorElement],
        value1: Union[IqoalaSingleton, IqoalaVectorElement],
        block_name: str,
    ) -> None:
        super().__init__(arguments=[value0, value1], attributes=[block_name])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have a result."
            )
        if len(args) != 2:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 2 arguments but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if not isinstance(attr, str):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation attribute must be a string."
            )
        if (
            not isinstance(args[0], IqoalaSingleton)
            and not isinstance(args[0], IqoalaVectorElement)
        ) or (
            not isinstance(args[1], IqoalaSingleton)
            and not isinstance(args[1], IqoalaVectorElement)
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )

        return cls(args[0], args[1], attr)


class BranchIfNotEqualOp(ClassicalIqoalaOp):
    OP_NAME = "bne"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        value0: Union[IqoalaSingleton, IqoalaVectorElement],
        value1: Union[IqoalaSingleton, IqoalaVectorElement],
        block_name: str,
    ) -> None:
        super().__init__(arguments=[value0, value1], attributes=[block_name])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have a result."
            )
        if len(args) != 2:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 2 arguments but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if not isinstance(attr, str):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation attribute must be a string."
            )
        if (
            not isinstance(args[0], IqoalaSingleton)
            and not isinstance(args[0], IqoalaVectorElement)
        ) or (
            not isinstance(args[1], IqoalaSingleton)
            and not isinstance(args[1], IqoalaVectorElement)
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )
        return cls(args[0], args[1], attr)


class BranchIfGreaterThanOp(ClassicalIqoalaOp):
    OP_NAME = "bgt"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        value0: Union[IqoalaSingleton, IqoalaVectorElement],
        value1: Union[IqoalaSingleton, IqoalaVectorElement],
        block_name: str,
    ) -> None:
        super().__init__(arguments=[value0, value1], attributes=[block_name])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have a result."
            )
        if len(args) != 2:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 2 arguments but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if not isinstance(attr, str):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation attribute must be a string."
            )
        if (
            not isinstance(args[0], IqoalaSingleton)
            and not isinstance(args[0], IqoalaVectorElement)
        ) or (
            not isinstance(args[1], IqoalaSingleton)
            and not isinstance(args[1], IqoalaVectorElement)
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )
        return cls(args[0], args[1], attr)


class BranchIfLessThanOp(ClassicalIqoalaOp):
    OP_NAME = "blt"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self,
        value0: Union[IqoalaSingleton, IqoalaVectorElement],
        value1: Union[IqoalaSingleton, IqoalaVectorElement],
        block_name: str,
    ) -> None:
        super().__init__(arguments=[value0, value1], attributes=[block_name])

    @classmethod
    def from_generic_args(
        cls,
        result: Optional[IqoalaVar],
        args: List[IqoalaVar],
        attr: Optional[IqoalaValue],
    ):
        if result is not None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation cannot have a result."
            )
        if len(args) != 2:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation takes 2 arguments but got {len(args)}."
            )
        if attr is None:
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation must have an attribute."
            )
        if not isinstance(attr, str):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation attribute must be a string."
            )
        if (
            not isinstance(args[0], IqoalaSingleton)
            and not isinstance(args[0], IqoalaVectorElement)
        ) or (
            not isinstance(args[1], IqoalaSingleton)
            and not isinstance(args[1], IqoalaVectorElement)
        ):
            raise HostLanguageSyntaxError(
                f"{cls.OP_NAME} operation arguments must be strings or vector elements."
            )
        return cls(args[0], args[1], attr)


class BasicBlockType(Enum):
    CL = 0
    CC = auto()
    QL = auto()
    QC = auto()


@dataclass
class BasicBlock:
    name: str
    typ: BasicBlockType
    instructions: List[ClassicalIqoalaOp]
    deadlines: Optional[Dict[str, int]] = None

    def __str__(self) -> str:
        annotations = f"type = {self.typ.name}"
        if self.deadlines is not None:
            annotations += f", deadlines: {self.deadlines}"
        annotations = "{" + annotations + "}"
        s = f"^{self.name} {annotations}:\n"
        return s + "\n".join("    " + str(i) for i in self.instructions)



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Optional

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    NvParams,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.from_nv_params(
            num_qubits=num_qubits, params=NvParams()
        ),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("NvNtf"),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text, flavour=NVFlavour()).parse()


@dataclass
class QkdResult:
    alice_result: BatchResult
    bob_result: BatchResult


def run_qkd(
    num_iterations: int,
    alice_file: str,
    bob_file: str,
    num_pairs: Optional[int] = None,
    linear: bool = True,
):
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )

    alice_program = load_program(alice_file)
    bob_program = load_program(bob_file)

    if num_pairs is not None:
        alice_input = ProgramInput({"bob_id": bob_id, "N": num_pairs})
        bob_input = ProgramInput({"alice_id": alice_id, "N": num_pairs})
    else:
        alice_input = ProgramInput({"bob_id": bob_id})
        bob_input = ProgramInput({"alice_id": alice_id})

    app_result = run_two_node_app(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_input, "bob": bob_input},
        network_cfg=network_cfg,
        linear=linear,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return QkdResult(alice_result, bob_result)


def qkd_1pair_md():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_1pair_MD_alice.iqoala"
    bob_file = "qkd_1pair_MD_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]


def qkd_1pair_ck():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_1pair_CK_alice.iqoala"
    bob_file = "qkd_1pair_CK_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]


def qkd_1pair_ck_cb():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_1pair_CK_cb_alice.iqoala"
    bob_file = "qkd_1pair_CK_cb_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]


def qkd_2pairs_md():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_2pairs_MD_alice.iqoala"
    bob_file = "qkd_2pairs_MD_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)

    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]
        assert alice["m1"] == bob["m1"]


def qkd_2pairs_ck_1qubit():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_2pairs_CK_1qubit_alice.iqoala"
    bob_file = "qkd_2pairs_CK_1qubit_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]
        assert alice["m1"] == bob["m1"]


def qkd_2pairs_ck_1qubit_cb():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_2pairs_CK_1qubit_cb_alice.iqoala"
    bob_file = "qkd_2pairs_CK_1qubit_cb_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file, linear=False)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]
        assert alice["m1"] == bob["m1"]


def qkd_2pairs_ck_2qubits_app_move():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_2pairs_CK_2qubits_app_move_alice_NV.iqoala"
    bob_file = "qkd_2pairs_CK_2qubits_app_move_bob_NV.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]
        assert alice["m1"] == bob["m1"]


def qkd_2pairs_ck_2qubits_wait_all():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_2pairs_CK_2qubits_wait_all_alice.iqoala"
    bob_file = "qkd_2pairs_CK_2qubits_wait_all_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]
        assert alice["m1"] == bob["m1"]


def qkd_100pairs_md():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_100pairs_MD_alice.iqoala"
    bob_file = "qkd_100pairs_MD_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)

    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        print(f"alice: {alice['outcomes']}")
        print(f"bob: {bob['outcomes']}")
        assert alice["outcomes"] == bob["outcomes"]


def qkd_npairs_md():
    ns.sim_reset()

    num_iterations = 1
    alice_file = "qkd_npairs_MD_alice.iqoala"
    bob_file = "qkd_npairs_MD_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file, num_pairs=100)

    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        print(f"alice: {alice['outcomes']}")
        print(f"bob  : {bob['outcomes']}")
        assert alice["outcomes"] == bob["outcomes"]


def qkd_npairs_ck_1qubit_cb():
    ns.sim_reset()

    num_iterations = 1
    alice_file = "qkd_npairs_CK_cb_alice.iqoala"
    bob_file = "qkd_npairs_CK_cb_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file, num_pairs=10)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        print(f"alice: {alice['outcomes']}")
        print(f"bob  : {bob['outcomes']}")
        assert alice["outcomes"] == bob["outcomes"]


def test_qkd_1pair_md():
    qkd_1pair_md()


def test_qkd_1pair_ck():
    qkd_1pair_ck()


def test_qkd_1pair_ck_cb():
    qkd_1pair_ck_cb()


def test_qkd_2pairs_md():
    qkd_2pairs_md()


def test_qkd_2pairs_ck_1qubit():
    qkd_2pairs_ck_1qubit()


def test_qkd_2pairs_ck_1qubit_cb():
    qkd_2pairs_ck_1qubit_cb()


def test_qkd_2pairs_ck_2qubits_app_move():
    qkd_2pairs_ck_2qubits_app_move()


def test_qkd_2pairs_ck_2qubits_wait_all():
    qkd_2pairs_ck_2qubits_wait_all()


def test_qkd_100pairs_md():
    qkd_100pairs_md()


def test_qkd_npairs_md():
    qkd_npairs_md()


def test_qkd_npairs_ck_1qubit_cb():
    qkd_npairs_ck_1qubit_cb()


if __name__ == "__main__":
    test_qkd_1pair_md()
    test_qkd_1pair_ck()
    test_qkd_1pair_ck_cb()
    test_qkd_2pairs_md()
    test_qkd_2pairs_ck_1qubit()
    test_qkd_2pairs_ck_1qubit_cb()
    test_qkd_2pairs_ck_2qubits_app_move()
    test_qkd_2pairs_ck_2qubits_wait_all()
    test_qkd_100pairs_md()
    test_qkd_npairs_md()
    test_qkd_npairs_ck_1qubit_cb()



from typing import Dict, Generator, List, Optional, Tuple, Type

import netsquid as ns
from netsquid.nodes import Node

from pydynaa import EventExpression
from qoala.lang.ehi import EhiBuilder, EhiNetworkInfo, UnitModule
from qoala.lang.program import ProgramMeta, QoalaProgram
from qoala.lang.request import (
    CallbackType,
    EprRole,
    EprType,
    QoalaRequest,
    RequestRoutine,
    RequestVirtIdMapping,
)
from qoala.lang.routine import LocalRoutine
from qoala.runtime.lhi import LhiLinkInfo, LhiTopologyBuilder
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.message import Message, RrCallTuple
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.sharedmem import MemAddr
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.entdist.entdist import EntDist, EntDistRequest
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.eprsocket import EprSocket
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack import NetstackInterface, NetstackLatencies
from qoala.sim.netstack.netstack import Netstack
from qoala.sim.netstack.netstackcomp import NetstackComponent
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice
from qoala.util.logging import LogManager
from qoala.util.math import B00_DENS, has_multi_state


class MockNetstackInterface(NetstackInterface):
    def __init__(
        self,
        comp: NetstackComponent,
        ehi_network: EhiNetworkInfo,
        qdevice: QDevice,
        requests: List[EntDistRequest],
    ) -> None:
        super().__init__(comp, ehi_network, qdevice, None)
        self._requests = requests


def create_n_qdevices(n: int, num_qubits: int = 1) -> List[QDevice]:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits)
    qdevices: List[QDevice] = []
    for i in range(n):
        qproc = build_qprocessor_from_topology(name=f"qproc_{i}", topology=topology)
        node = Node(name=f"node_{i}", qmemory=qproc)
        qdevices.append(QDevice(node=node, topology=topology))

    return qdevices


def create_alice_bob_qdevices(
    num_qubits: int = 1, alice_id: int = 0, bob_id: int = 1
) -> Tuple[QDevice, QDevice]:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits)

    alice_qproc = build_qprocessor_from_topology(name="qproc_alice", topology=topology)
    bob_qproc = build_qprocessor_from_topology(name="qproc_bob", topology=topology)

    alice_node = Node(name="alice", qmemory=alice_qproc, ID=alice_id)
    bob_node = Node(name="bob", qmemory=bob_qproc, ID=bob_id)

    alice_qdevice = QDevice(node=alice_node, topology=topology)
    bob_qdevice = QDevice(node=bob_node, topology=topology)

    return alice_qdevice, bob_qdevice


def create_request(
    node1_id: int,
    node2_id: int,
    local_qubit_id: int = 0,
    lpid: Optional[int] = 0,
    rpid: Optional[int] = 0,
) -> EntDistRequest:
    return EntDistRequest(
        local_node_id=node1_id,
        remote_node_id=node2_id,
        local_qubit_id=local_qubit_id,
        local_pid=lpid,
        remote_pid=rpid,
    )


def setup_components() -> Tuple[
    NetstackComponent, QDevice, NetstackComponent, QDevice, EntDist
]:
    alice, bob = create_alice_bob_qdevices(num_qubits=3)

    ehi_network = EhiNetworkInfo.only_nodes(
        {alice.node.ID: alice.node.name, bob.node.ID: bob.node.name}
    )
    alice_comp = NetstackComponent(alice.node, ehi_network)
    bob_comp = NetstackComponent(bob.node, ehi_network)
    entdist_comp = EntDistComponent(ehi_network)

    entdist = EntDist(
        nodes=[alice.node, bob.node], ehi_network=ehi_network, comp=entdist_comp
    )

    alice_comp.entdist_out_port.connect(entdist_comp.node_in_port("alice"))
    alice_comp.entdist_in_port.connect(entdist_comp.node_out_port("alice"))
    bob_comp.entdist_out_port.connect(entdist_comp.node_in_port("bob"))
    bob_comp.entdist_in_port.connect(entdist_comp.node_out_port("bob"))

    link_info = LhiLinkInfo.perfect(1000)
    entdist.add_sampler(alice.node.ID, bob.node.ID, link_info)

    return alice_comp, alice, bob_comp, bob, entdist


def test_single_pair_only_netstack_interface():
    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            yield from self.wait(500)
            self.send_entdist_msg(Message(0, 0, self._requests[0]))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            yield from self.wait(800)
            self.send_entdist_msg(Message(0, 0, self._requests[0]))

    alice_comp, alice_qdevice, bob_comp, bob_qdevice, entdist = setup_components()
    ehi_network: EhiNetworkInfo = entdist._ehi_network
    alice_id = alice_comp.node.ID
    bob_id = bob_comp.node.ID

    request_alice = create_request(alice_id, bob_id, 0, 0, 0)
    request_bob = create_request(bob_id, alice_id, 0, 0, 0)

    alice_intf = AliceNetstackInterface(
        alice_comp, ehi_network, alice_qdevice, requests=[request_alice]
    )
    bob_intf = BobNetstackInterface(
        bob_comp, ehi_network, bob_qdevice, requests=[request_bob]
    )
    ns.sim_reset()
    assert ns.sim_time() == 0

    alice_intf.start()
    bob_intf.start()
    entdist.start()
    ns.sim_run()

    assert ns.sim_time() == 1800  # 1000 + max(800, 500)

    alice_qubit = alice_qdevice.get_local_qubit(0)
    bob_qubit = bob_qdevice.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)


def test_multiple_pairs_only_netstack_interface():
    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            for request in self._requests:
                yield from self.wait(500)
                self.send_entdist_msg(Message(0, 0, request))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            for request in self._requests:
                yield from self.wait(500)
                self.send_entdist_msg(Message(0, 0, request))

    alice_comp, alice_qdevice, bob_comp, bob_qdevice, entdist = setup_components()
    ehi_network: EhiNetworkInfo = entdist._ehi_network
    alice_id = alice_comp.node.ID
    bob_id = bob_comp.node.ID

    requests_alice = [
        create_request(alice_id, bob_id, 0, 0, 0),
        create_request(alice_id, bob_id, 1, 0, 0),
        create_request(alice_id, bob_id, 2, 0, 0),
    ]
    requests_bob = [
        create_request(bob_id, alice_id, 1, 0, 0),
        create_request(bob_id, alice_id, 2, 0, 0),
        create_request(bob_id, alice_id, 0, 0, 0),
    ]

    alice_intf = AliceNetstackInterface(
        alice_comp, ehi_network, alice_qdevice, requests=requests_alice
    )
    bob_intf = BobNetstackInterface(
        bob_comp, ehi_network, bob_qdevice, requests=requests_bob
    )

    ns.sim_reset()
    assert ns.sim_time() == 0
    alice_intf.start()
    bob_intf.start()
    entdist.start()
    ns.sim_run()

    # 500 + 1000 * 3 since epr pairs are created sequentially
    assert ns.sim_time() == 3500

    alice_q0, alice_q1, alice_q2 = [alice_qdevice.get_local_qubit(i) for i in range(3)]
    bob_q0, bob_q1, bob_q2 = [bob_qdevice.get_local_qubit(i) for i in range(3)]
    assert has_multi_state([alice_q0, bob_q1], B00_DENS)
    assert has_multi_state([alice_q1, bob_q2], B00_DENS)
    assert has_multi_state([alice_q2, bob_q0], B00_DENS)

    assert not has_multi_state([alice_q0, bob_q0], B00_DENS)
    assert not has_multi_state([alice_q1, bob_q1], B00_DENS)
    assert not has_multi_state([alice_q2, bob_q2], B00_DENS)
    assert not has_multi_state([alice_q0, bob_q2], B00_DENS)
    assert not has_multi_state([alice_q1, bob_q0], B00_DENS)
    assert not has_multi_state([alice_q2, bob_q1], B00_DENS)


def setup_components_full_netstack(
    num_qubits: int,
    alice_id: int,
    bob_id: int,
    alice_netstack_cls: Type[Netstack],
    bob_netstack_cls: Type[Netstack],
) -> Tuple[Netstack, Netstack, EntDist]:
    alice_qdevice, bob_qdevice = create_alice_bob_qdevices(
        num_qubits=num_qubits, alice_id=alice_id, bob_id=bob_id
    )

    ehi_network = EhiNetworkInfo.only_nodes(
        {
            alice_qdevice.node.ID: alice_qdevice.node.name,
            bob_qdevice.node.ID: bob_qdevice.node.name,
        }
    )
    alice_comp = NetstackComponent(alice_qdevice.node, ehi_network)
    bob_comp = NetstackComponent(bob_qdevice.node, ehi_network)
    entdist_comp = EntDistComponent(ehi_network)

    entdist = EntDist(
        nodes=[alice_qdevice.node, bob_qdevice.node],
        ehi_network=ehi_network,
        comp=entdist_comp,
    )

    alice_comp.entdist_out_port.connect(entdist_comp.node_in_port("alice"))
    alice_comp.entdist_in_port.connect(entdist_comp.node_out_port("alice"))
    bob_comp.entdist_out_port.connect(entdist_comp.node_in_port("bob"))
    bob_comp.entdist_in_port.connect(entdist_comp.node_out_port("bob"))

    link_info = LhiLinkInfo.perfect(1000)
    entdist.add_sampler(alice_qdevice.node.ID, bob_qdevice.node.ID, link_info)

    alice_netstack = alice_netstack_cls(
        comp=alice_comp,
        ehi_network=ehi_network,
        memmgr=MemoryManager("alice", alice_qdevice),
        qdevice=alice_qdevice,
        latencies=NetstackLatencies.all_zero(),
    )
    bob_netstack = bob_netstack_cls(
        comp=bob_comp,
        ehi_network=ehi_network,
        memmgr=MemoryManager("bob", bob_qdevice),
        qdevice=bob_qdevice,
        latencies=NetstackLatencies.all_zero(),
    )

    return alice_netstack, bob_netstack, entdist


def test_single_pair_full_netstack():
    alice_id = 0
    bob_id = 1

    request_alice = create_request(alice_id, bob_id, 0, 0, 0)
    request_bob = create_request(bob_id, alice_id, 0, 0, 0)

    class AliceNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            yield from self.processor._execute_entdist_request(request_alice)

    class BobNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            yield from self.processor._execute_entdist_request(request_bob)

    alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
        1, alice_id, bob_id, AliceNetstack, BobNetstack
    )

    ns.sim_reset()
    assert ns.sim_time() == 0
    alice_netstack.start()
    bob_netstack.start()
    entdist.start()
    ns.sim_run()

    assert ns.sim_time() == 1000

    alice_qubit = alice_netstack.qdevice.get_local_qubit(0)
    bob_qubit = bob_netstack.qdevice.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)


def test_multiple_pairs_full_netstack():
    ns.sim_reset()

    alice_id = 0
    bob_id = 1

    requests_alice = [
        create_request(alice_id, bob_id, 0, 0, 0),
        create_request(alice_id, bob_id, 1, 1, 1),
    ]
    requests_bob = [
        create_request(bob_id, alice_id, 0, 0, 0),
        create_request(bob_id, alice_id, 2, 1, 1),
    ]

    class AliceNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            for request in requests_alice:
                yield from self.processor._execute_entdist_request(request)

    class BobNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            for request in requests_bob:
                yield from self.processor._execute_entdist_request(request)

    alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
        3, alice_id, bob_id, AliceNetstack, BobNetstack
    )

    alice_netstack.start()
    bob_netstack.start()
    entdist.start()
    ns.sim_run()

    aq0 = alice_netstack.qdevice.get_local_qubit(0)
    bq0 = bob_netstack.qdevice.get_local_qubit(0)
    aq1 = alice_netstack.qdevice.get_local_qubit(1)
    bq2 = bob_netstack.qdevice.get_local_qubit(2)
    assert has_multi_state([aq0, bq0], B00_DENS)
    assert has_multi_state([aq1, bq2], B00_DENS)


def create_simple_request(
    remote_id: int,
    num_pairs: int,
    virt_ids: RequestVirtIdMapping,
    typ: EprType,
    role: EprRole,
) -> QoalaRequest:
    return QoalaRequest(
        name="req",
        remote_id=remote_id,
        epr_socket_id=0,
        num_pairs=num_pairs,
        virt_ids=virt_ids,
        timeout=1000,
        fidelity=0.65,
        typ=typ,
        role=role,
    )


def simple_req_routine(
    remote_id: int,
    num_pairs: int,
    virt_ids: RequestVirtIdMapping,
    typ: EprType,
    role: EprRole,
) -> RequestRoutine:
    return RequestRoutine(
        name="req1",
        request=create_simple_request(remote_id, num_pairs, virt_ids, typ, role),
        return_vars=[],
        callback_type=CallbackType.WAIT_ALL,
        callback=None,
    )


def create_process(
    num_qubits: int,
    routines: Optional[Dict[str, LocalRoutine]] = None,
    req_routines: Optional[Dict[str, RequestRoutine]] = None,
    epr_sockets: Optional[Dict[int, EprSocket]] = None,
) -> QoalaProcess:
    if routines is None:
        routines = {}
    if req_routines is None:
        req_routines = {}
    if epr_sockets is None:
        epr_sockets = {}
    program = QoalaProgram(
        blocks=[],
        local_routines=routines,
        meta=ProgramMeta.empty(""),
        request_routines=req_routines,
    )

    ehi = EhiBuilder.perfect_uniform(num_qubits, None, [], 0, [], 0)
    unit_module = UnitModule.from_full_ehi(ehi)
    instance = ProgramInstance(
        pid=0,
        program=program,
        inputs=ProgramInput({}),
        unit_module=unit_module,
    )
    mem = ProgramMemory(pid=0)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def test_single_pair_qoala_ck_request_only_alice():
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    routine_alice = simple_req_routine(
        remote_id=bob_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.CREATE_KEEP,
        role=EprRole.CREATE,
    )

    requests_bob = [
        create_request(bob_id, alice_id, 0, 0, 0),
        create_request(bob_id, alice_id, 1, 0, 0),
    ]
    epr_socket = EprSocket(0, bob_id, 0, 0, 1.0)
    process_alice = create_process(
        num_qubits, req_routines={"req1": routine_alice}, epr_sockets={0: epr_socket}
    )

    class AliceNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            rrcall = RrCallTuple.no_alloc("req1")
            yield from self.processor.assign_request_routine(process_alice, rrcall)

    class BobNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            for request in requests_bob:
                yield from self.processor._execute_entdist_request(request)

    alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
        num_qubits, alice_id, bob_id, AliceNetstack, BobNetstack
    )
    alice_netstack.interface.memmgr.add_process(process_alice)

    alice_netstack.start()
    bob_netstack.start()
    entdist.start()
    ns.sim_run()

    aq0 = alice_netstack.qdevice.get_local_qubit(0)
    bq0 = bob_netstack.qdevice.get_local_qubit(0)
    aq1 = alice_netstack.qdevice.get_local_qubit(1)
    bq1 = bob_netstack.qdevice.get_local_qubit(1)
    assert has_multi_state([aq0, bq0], B00_DENS)
    assert has_multi_state([aq1, bq1], B00_DENS)


def test_single_pair_qoala_ck_request():
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    routine_alice = simple_req_routine(
        remote_id=bob_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.CREATE_KEEP,
        role=EprRole.CREATE,
    )
    routine_bob = simple_req_routine(
        remote_id=alice_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.CREATE_KEEP,
        role=EprRole.RECEIVE,
    )

    epr_socket_alice = EprSocket(0, bob_id, 0, 0, 1.0)
    process_alice = create_process(
        num_qubits,
        req_routines={"req1": routine_alice},
        epr_sockets={0: epr_socket_alice},
    )
    epr_socket_bob = EprSocket(0, alice_id, 0, 0, 1.0)
    process_bob = create_process(
        num_qubits, req_routines={"req1": routine_bob}, epr_sockets={0: epr_socket_bob}
    )

    class AliceNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            rrcall = RrCallTuple.no_alloc("req1")
            yield from self.processor.assign_request_routine(process_alice, rrcall)

    class BobNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            rrcall = RrCallTuple.no_alloc("req1")
            yield from self.processor.assign_request_routine(process_bob, rrcall)

    alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
        num_qubits, alice_id, bob_id, AliceNetstack, BobNetstack
    )
    alice_netstack.interface.memmgr.add_process(process_alice)
    bob_netstack.interface.memmgr.add_process(process_bob)

    alice_netstack.start()
    bob_netstack.start()
    entdist.start()
    ns.sim_run()

    aq0 = alice_netstack.qdevice.get_local_qubit(0)
    bq0 = bob_netstack.qdevice.get_local_qubit(0)
    aq1 = alice_netstack.qdevice.get_local_qubit(1)
    bq1 = bob_netstack.qdevice.get_local_qubit(1)
    assert has_multi_state([aq0, bq0], B00_DENS)
    assert has_multi_state([aq1, bq1], B00_DENS)


def test_single_pair_qoala_ck_custom_virt_ids():
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    routine_alice = simple_req_routine(
        remote_id=bob_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("custom 2, 0"),
        typ=EprType.CREATE_KEEP,
        role=EprRole.CREATE,
    )
    routine_bob = simple_req_routine(
        remote_id=alice_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("custom 1, 2"),
        typ=EprType.CREATE_KEEP,
        role=EprRole.RECEIVE,
    )

    epr_socket_alice = EprSocket(0, bob_id, 0, 0, 1.0)
    process_alice = create_process(
        num_qubits,
        req_routines={"req1": routine_alice},
        epr_sockets={0: epr_socket_alice},
    )
    epr_socket_bob = EprSocket(0, alice_id, 0, 0, 1.0)
    process_bob = create_process(
        num_qubits, req_routines={"req1": routine_bob}, epr_sockets={0: epr_socket_bob}
    )

    class AliceNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            rrcall = RrCallTuple.no_alloc("req1")
            yield from self.processor.assign_request_routine(process_alice, rrcall)

    class BobNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            rrcall = RrCallTuple.no_alloc("req1")
            yield from self.processor.assign_request_routine(process_bob, rrcall)

    alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
        num_qubits, alice_id, bob_id, AliceNetstack, BobNetstack
    )
    alice_netstack.interface.memmgr.add_process(process_alice)
    bob_netstack.interface.memmgr.add_process(process_bob)

    alice_netstack.start()
    bob_netstack.start()
    entdist.start()
    ns.sim_run()

    alice_phys_id = alice_netstack._interface._memmgr.phys_id_for(process_alice.pid, 2)
    alice_phys_id2 = alice_netstack._interface._memmgr.phys_id_for(process_alice.pid, 0)

    bob_phys_id = bob_netstack._interface._memmgr.phys_id_for(process_bob.pid, 1)
    bob_phys_id2 = bob_netstack._interface._memmgr.phys_id_for(process_bob.pid, 2)

    aq = alice_netstack.qdevice.get_local_qubit(alice_phys_id)
    aq2 = alice_netstack.qdevice.get_local_qubit(alice_phys_id2)

    bq = bob_netstack.qdevice.get_local_qubit(bob_phys_id)
    bq2 = bob_netstack.qdevice.get_local_qubit(bob_phys_id2)
    assert has_multi_state([aq, bq], B00_DENS)
    assert has_multi_state([aq2, bq2], B00_DENS)


def test_single_pair_qoala_md_request_different_virt_ids():
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    routine_alice = simple_req_routine(
        remote_id=bob_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.CREATE,
    )
    routine_bob = simple_req_routine(
        remote_id=alice_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.RECEIVE,
    )

    epr_socket_alice = EprSocket(0, bob_id, 0, 0, 1.0)
    process_alice = create_process(
        num_qubits,
        req_routines={"req1": routine_alice},
        epr_sockets={0: epr_socket_alice},
    )
    epr_socket_bob = EprSocket(0, alice_id, 0, 0, 1.0)
    process_bob = create_process(
        num_qubits, req_routines={"req1": routine_bob}, epr_sockets={0: epr_socket_bob}
    )

    class AliceNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            shared_mem = process_alice.prog_memory.shared_mem
            result_addr = shared_mem.allocate_rr_out(2)
            rrcall = RrCallTuple(
                "req1",
                input_addr=MemAddr(0),
                result_addr=result_addr,
                cb_input_addrs=[],
                cb_output_addrs=[],
            )
            self.outcomes = yield from self.processor.assign_request_routine(
                process_alice, rrcall
            )

    class BobNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            shared_mem = process_bob.prog_memory.shared_mem
            result_addr = shared_mem.allocate_rr_out(2)
            rrcall = RrCallTuple(
                "req1",
                input_addr=MemAddr(0),
                result_addr=result_addr,
                cb_input_addrs=[],
                cb_output_addrs=[],
            )
            self.outcomes = yield from self.processor.assign_request_routine(
                process_bob, rrcall
            )

    alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
        num_qubits, alice_id, bob_id, AliceNetstack, BobNetstack
    )
    alice_netstack.interface.memmgr.add_process(process_alice)
    bob_netstack.interface.memmgr.add_process(process_bob)

    alice_netstack.start()
    bob_netstack.start()
    entdist.start()
    ns.sim_run()

    # All virtual qubits should be free.
    assert alice_netstack.interface.memmgr.phys_id_for(process_alice.pid, 0) is None
    assert alice_netstack.interface.memmgr.phys_id_for(process_alice.pid, 1) is None
    assert bob_netstack.interface.memmgr.phys_id_for(process_bob.pid, 0) is None
    assert bob_netstack.interface.memmgr.phys_id_for(process_bob.pid, 1) is None

    assert alice_netstack.outcomes == bob_netstack.outcomes


def test_single_pair_qoala_md_request_same_virt_ids():
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    routine_alice = simple_req_routine(
        remote_id=bob_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("all 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.CREATE,
    )
    routine_bob = simple_req_routine(
        remote_id=alice_id,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("all 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.RECEIVE,
    )

    epr_socket_alice = EprSocket(0, bob_id, 0, 0, 1.0)
    process_alice = create_process(
        num_qubits,
        req_routines={"req1": routine_alice},
        epr_sockets={0: epr_socket_alice},
    )
    epr_socket_bob = EprSocket(0, alice_id, 0, 0, 1.0)
    process_bob = create_process(
        num_qubits, req_routines={"req1": routine_bob}, epr_sockets={0: epr_socket_bob}
    )

    class AliceNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            shared_mem = process_alice.prog_memory.shared_mem
            self.result_addr = shared_mem.allocate_rr_out(2)
            rrcall = RrCallTuple(
                "req1",
                input_addr=MemAddr(0),
                result_addr=self.result_addr,
                cb_input_addrs=[],
                cb_output_addrs=[],
            )
            self.outcomes = yield from self.processor.assign_request_routine(
                process_alice, rrcall
            )

    class BobNetstack(Netstack):
        def run(self) -> Generator[EventExpression, None, None]:
            shared_mem = process_bob.prog_memory.shared_mem
            self.result_addr = shared_mem.allocate_rr_out(2)
            rrcall = RrCallTuple(
                "req1",
                input_addr=MemAddr(0),
                result_addr=self.result_addr,
                cb_input_addrs=[],
                cb_output_addrs=[],
            )
            self.outcomes = yield from self.processor.assign_request_routine(
                process_bob, rrcall
            )

    alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
        num_qubits, alice_id, bob_id, AliceNetstack, BobNetstack
    )
    alice_netstack.interface.memmgr.add_process(process_alice)
    bob_netstack.interface.memmgr.add_process(process_bob)

    alice_netstack.start()
    bob_netstack.start()
    entdist.start()
    ns.sim_run()

    # All virtual qubits should be free.
    assert alice_netstack.interface.memmgr.phys_id_for(process_alice.pid, 0) is None
    assert alice_netstack.interface.memmgr.phys_id_for(process_alice.pid, 1) is None
    assert bob_netstack.interface.memmgr.phys_id_for(process_bob.pid, 0) is None
    assert bob_netstack.interface.memmgr.phys_id_for(process_bob.pid, 1) is None

    alice_result = process_alice.shared_mem.read_rr_out(alice_netstack.result_addr, 2)
    bob_result = process_bob.shared_mem.read_rr_out(bob_netstack.result_addr, 2)

    assert alice_result == bob_result


if __name__ == "__main__":
    # test_single_pair_only_netstack_interface()
    LogManager.set_log_level("INFO")
    test_multiple_pairs_only_netstack_interface()
    # test_single_pair_full_netstack()
    # test_multiple_pairs_full_netstack()
    # test_single_pair_qoala_ck_request_only_alice()
    # test_single_pair_qoala_ck_request()
    # test_single_pair_qoala_md_request_different_virt_ids()
    # test_single_pair_qoala_md_request_same_virt_ids()



import numpy as np

from qoala.util.math import density_matrices_equal


def test_1():
    state1 = np.array([[1, 0], [0, 0]])
    state2 = np.array([[0.99999999, 0], [0, 0]])
    assert density_matrices_equal(state1, state2)

    state1 = np.array([[1, 0], [0, 0]])
    state2 = np.array([[0.99999999, 0], [0, 0.1]])
    assert not density_matrices_equal(state1, state2)

    state1 = np.array([[1, 0], [0, 0]])
    state2 = np.array([[0, 0], [0, 1]])
    assert not density_matrices_equal(state1, state2)


if __name__ == "__main__":
    test_1()
